/*
This SAS program is meant to demonstrate the methods given in 
	Madden, L. V., Paul, P. A., and Lipps, P. E. 2007.
	Consideration of nonparametric approaches for assessing 
	genotype-by-environment (G x E) interactions with disease 
	severity data. Plant Disease 91: 891-900.
The approach, originally developed by Piepho, based on other
work on nonparametric methods, is described in this cited article.
Piepho and other references are given in the Madden et al. article.

/*
Friedman test and Kendall concordance, plus Piepho's treatment stability analysis.
Assumes that:
	blk is the location (=enviroment) factor,
	trt is the genotype (or any treatment) factor, and
	x is the response variable.
*/

options ls=80;
%macro interactE (data= , blk= , trt= , response = , noblks =, nocols= , notrts = ,
		print=no,spearman=no,odd=);

/*

Macro to determine statistics on concordance of ranks and also Piepho's 
stability measure of interaction. The macro calculates some statistics
in more than one way, to demonstrate different approaches for obtaining 
the same results. 

		Citation: Madden et al. 2007. PLANT DISEASE 91: 891-900.

The data file must have three columns: response variable ('response', 
x in the paper), a treatment (variety) code ('trt'), and an environment 
code ('blk'). The coding variables MUST have use only the sequence of 
integers (1, 2, 3, ...), with no skips, for treatments and environments.
In other words, one cannot use letters or names (A, B, check,...,etc.) for
either treatment or environment. If you are using names or letters to 
indicate treatment (variety) or environment, then you must 
first add variables to your data file consisting of only integers. 
Note: there can be no missing values for the response. 
 
Macro arguments:
data	Name of data file
blk		Name of the environment (location,block) coding variable
		(must be integers from 1 to noblks)
trt		Name of the treatment (variety, cultivar, etc.) coding variable
		(must be numbers from 1 to notrts)
response Name of the response (dependent) variable
noblks	Number of environments
nocols	Number of "new" environments (either noblks/2 or noblks/2 + 1)
		(value depends on even or odd number of environments)
notrts	Number of treatments (varieties, etc.)
print	Flag (no, yes) to indicate whether or not multiple tables of output 
		are diplayed (default is no). With print=yes, one sees more
		of the intermediate results, and tables of responses or ranks.
spearman Flag (no, yes) to indicate whether the matrix of pairwise
		Spearman rank correlations is displayed (default is no)
odd		Flag (no, yes) to indicate whether there is an odd or even number
		of environments. 

All arguments are required except for print and spearman.

Macro uses Title2-Title5 for explaining output. If you previously had 
titles specified, these will be overwritten. These are all blank at
the end of the macro.

*/

title2 'Obtaining Friedman Q (chi-square) and Kendall W ';


proc sort data=&data out=_a;
	by &blk;
proc rank data=_a out=_a;
	by &blk;
	var &response; ranks rx;
proc means data=_a noprint;
	class &trt;
	var rx;
	output mean(rx)=mean_rank var(rx)=StwoRX
		out=_a_out;
proc means data=_a noprint;
	class &trt;
	var &response;
	output	median(&response)=median_response 
			mean(&response)=mean_response 
			out=_a_out2;
data _a_; merge _a_out _a_out2;
	label median_response='Median Response' 
	mean_rank='Mean Rank'
	mean_response='Mean Response' 
	StwoRX='S(2)_i';
/*
proc sort data=_a out=_asort;
by &trt;
proc means data=_asort var mean;
by &trt;
var rx;
*/
*---One can obtain relevant statistics for Kendall statistic using 
	either proc anova (with the right form or ranks) or proc freq;
/*
proc anova data=_a;
	title2 'Friedman from PROC ANOVA; W = SSTrt/(SSTrt+SSError); Q = W*(#blks*(#trts-1))';
	class &blk &trt;
	model rx = &blk &trt;
	means &trt;
*/
	
proc freq data=_a;
	title3 'Friedman with PROC FREQ; Q is direct from Statistic 2 (Row Means Differ)';
	title4 'use blk*trt*Response, with CMH2 SCORES=RANK';
	tables &blk*&trt*&Response  / cmh2 scores=rank noprint;
	ods output cmh=cmh;
data cmh; set cmh;	/*Manipulations from above output to get needed statistics*/
	if (Statistic ne 2) then delete;
	W = Value/(&noblks*(&notrts-1));
	Q=Value;
	Fstat=(&noblks-1)*W/(1.0-W);
	df_F_=(&noblks-1)*(&notrts-1);

	df_NF = (&notrts-1) - (2/&noblks);
	df_DF = df_NF*(&noblks-1); 
	Pr_F=1-probF(Fstat,df_NF,df_DF);
	SumSi2=(1-W)*&noblks*&notrts*((&notrts**2)-1)/(12.0*(&noblks-1));
	if (W gt 0) then V_r_g=(1.0-W)/W; else V_r_g = 10**6;
	c_crit=cinv(0.95,DF);
		lam1=c_crit/(&noblks*(&notrts-1)-c_crit);
	Si2crit=(&notrts*&noblks*((&notrts**2)-1))/(12.*(&noblks-1)*(1+lam1));

	SSS = W*((&noblks**2)*&notrts*(&notrts**2 -1)/12.0);
	SSSmax= ((&noblks**2)*&notrts*(&notrts**2 -1)/12.0);

	label Q='Friedman chi-square' W='Kendall Coef. of Concordance'
		Fstat='Friedman F statistic' df_F_='Simple DDF for F' V_r_g=
		'Pred.Var.Ratio' SumSi2='SS for ranks' Si2crit='Critical SS for sig.'
		SSS='SS Rank Totals' SSSmax='Max SS RnkTot'
		df_NF='Num.DF for F' df_DF='Den.Df for F';
	Rows=&notrts; Columns=&noblks;

proc print data=cmh label;
	var Q DF Prob Fstat df_F_ df_NF df_DF Pr_F W V_r_g SumSi2 Si2crit lam1 c_crit
		SSS SSSmax Rows Columns;
proc print data=_a_ label;
/*
data _a_omit; set _a_;
if (&trt eq .) then delete;
proc means data=_a_omit sum;
run;
*/
title3 ; title4 ;

/*
Now transpose data, so that pairwise contrasts can be obtained for
Piepho analysis. Note: if #blks is odd, use contrast of blks 1-3.

This version is for even/odd number of blocks.

*/
title2 'Piepho stability analysis, based on column differences';

proc sort data=_a out=_a;
	by &trt;

proc transpose data=_a out=flipped;
	by &trt; id &blk;
	var &Response;

proc transpose data=_a out=flipp_r;
	by &trt; id &blk;
	var rx;

%if (&print ne no) %then %do;
	title2 'Unstacked data (K*N)';
	proc print data=flipped;
	proc print data=flipp_r;
	%end;
	title2 ;

%if (&spearman ne no) %then %do;
	proc corr data=flipped outs=_corrs spearman noprint;
		var _1 -- _&noblks;
	data _corrs; set _corrs; if _TYPE_ = 'CORR';
	title2 'Spearman rank correlations between environments';
	proc print data=_corrs round;
	%end;
	title2 ;
run;

title3 'Unstacked data, differences, and ranks';
/* The next part is for an ODD number of environments */

%if (%UPCASE(&odd) ne NO) %then %do;

data flipped; set flipped;
	%let nocol = &nocols;
	array temp{&noblks} _1 -- _&noblks;
	array new{&nocol} __1 - __&nocol;
		new{1} = temp{1}+temp{2} -2*temp{3};
		*new{1} = 2*temp{3} - temp{1} - temp{2} ;
	do i = 2 to &nocol;
		new{i}=temp{2*i+1}-temp{2*i};
		
	end;
*proc print data=flipped;
%end;

/*--now, if even....*/
%if (%UPCASE(&odd) eq NO) %then %do;
data flipped; set flipped;
	
	%let nocol = &nocols;

	array temp{&noblks} _1 -- _&noblks;
	array new{&nocol} __1 - __&nocol;
	*new{1} = temp{1}+temp{2} -2*temp{3};
	do i = 1 to &nocol;
		new{i}=temp{2*i}-temp{2*i-1};
		 
	end;
*proc print data=flipped;
%end;
	

proc rank data=flipped out=flipped;
	var __1 -- __&nocols; ranks _r1 - _r&nocols;

%if (&print ne no) %then %do;
	proc print data=flipped;
	%end;
	
/* put ranks back in stacked format for analyses. */

proc transpose data=flipped out=flipped2;
	by &trt; id _name_;
	var _r1 -- _r&nocols;
*proc print data=flipped;
data flipped2; set flipped2;
	*--Note, x or &Response now is the rank (not original) in this new file;
	w = (&Response -((&notrts+1)/2))**2;

	title2 'Data file with _NAME_ for new COLUMN (block) label, and W for new SCORE variable';
%if (&print ne no) %then %do;
	proc print data=flipped2;
proc means data=flipped2 sum noprint;
	class &trt;
	var w &Response;
	output out=flipp2out sum(w)=wsum mean(&Response)=mnrank;
data flipp2out; set flipp2out; 
	if (&trt eq .) then delete;
	Ui=wsum/&nocols;
	Uisqu=Ui**2;
proc print data=flipp2out;
	var &trt Ui Uisqu mnrank;
proc means data=flipp2out sum;
	var Uisqu;

%end;

run;

*---After a lot of manipulation (above), one can now get the needed 
	results by using either proc anova or proc freq with the Piepho
	SCORE-statistic variable (=w, rank of column differences).;
/*
proc anova data=flipped2;
	title3 'Test with PROC ANOVA; use W score for response (~rank of column difference)';
	title4 'Q = (SSTrt/[SSTrt+SSError])*(#columns*(#trts-1))';
	class &trt _name_;
	model w = _name_ &trt;
*/

proc freq data=flipped2;
	title3 'Test with PROC FREQ; use _NAME_*TRT*W, with CMH2';
	title4 'W is score for response (~~rank of column difference)';
	title5 'Q is from Statistic 2 (Row Means Differ)';
	tables _name_*&trt*w / cmh2 noprint;
	ods output cmh=cmh2;

data cmh2; set cmh2 ;	
*---A lot more manipulation to get test statistics and different forms of d.f.;

if (Statistic ne 2) then delete;

	Columns=&nocols; Rows=&notrts;
	W = Value/(&nocols*(&notrts-1));
	U=Value;

	a00 = (5*&nocols*(&notrts**2 - 1)*(&notrts - 1)) / (4*(&notrts**2 - 4));
	a01 = 180*&nocols/(&notrts*(&notrts+1)*(&notrts**2 - 4));
	sumUi2 = (U + a00) / a01;
	if (&nocols gt 2) then do;
		a = &nocols*(&nocols-1)*(&notrts-1)/((&nocols-2)**2);
		b = &nocols/(&nocols-2);
	end;
	if (&nocols gt 2) then 	Up = (a/(&nocols-1)) + b*U;
		else Up = U;
	if (&nocols gt 2) then	Pr_Up = 1 - probchi(Up,a);
		else Pr_Up = Prob;
	Fstat=(&nocols-1)*W/(1.0-W);
	df_F = (&nocols-1)*(&notrts-1);
	Pr_F=1-probF(Fstat,&notrts-1,(&nocols-1)*(&notrts-1));

	dfNF = (&notrts-1) - (2/&nocols);
	dfDF = dfNF*(&nocols-1);
	Pr_F2 = 1-probF(Fstat,dfNF,dfDF);

	label Value='Chi-square Q diff' W='Kend.Coef. Conc. Diff.' U='U statistic (diff.)'
	Fstat='F statistic (diff.)' df_F='Den. df for F (diff.)' Pr_F='Prob. F (diff.)'
	Up='Modified U statistic (U_p)' Pr_Up='Prob. U_p'
	dfNF='improvedNDF' dfDF='improvedDDF' Pr_F2='improvedProbF'
	simUi2 ='sum(U_i)2';
	;

*---There are different possible test statistics and different ways of getting
	numerator and denominator d.f. for the analysis. Output gives several sets
	of results.;

proc print data=cmh2 label;
	var Value df Prob W U sumUi2 a b Up Pr_Up Fstat df_F Pr_F 
	dfNF dfDF Pr_F2 Rows Columns;
run;
title2 ; title3 ; title4 ; title5 ;
run;
%mend;



data s96;	*---Southern 1996 data for Gray Leaf Spot of Corn (11 locations);
input study x genChar $ gen loc;  /*genotype indicated with names _and_ numbers*/
datalines;
961	27	P3394	1	1
961	36	FR1064xLH185	2	1
961	17	P3163	3	1
961	15	LH195xLH212	4	1
961	7	LH195xLH210	5	1
961	17	LH235xLH185	6	1
961	14	LH200xLH216	7	1
961	3	LH200xLH172	8	1
961	2	LH195xNC258	9	1
961	3	DK683	10	1
961	22	P3394	1	2
961	21	FR1064xLH185	2	2
961	14	P3163	3	2
961	8	LH195xLH212	4	2
961	8	LH195xLH210	5	2
961	7	LH235xLH185	6	2
961	4	LH200xLH216	7	2
961	4	LH200xLH172	8	2
961	2	LH195xNC258	9	2
961	3	DK683	10	2
961	35	P3394	1	3
961	43	FR1064xLH185	2	3
961	20	P3163	3	3
961	21	LH195xLH212	4	3
961	23	LH195xLH210	5	3
961	25	LH235xLH185	6	3
961	19	LH200xLH216	7	3
961	19	LH200xLH172	8	3
961	11	LH195xNC258	9	3
961	5	DK683	10	3
961	30	P3394	1	4
961	23	FR1064xLH185	2	4
961	14	P3163	3	4
961	13	LH195xLH212	4	4
961	11	LH195xLH210	5	4
961	13	LH235xLH185	6	4
961	11	LH200xLH216	7	4
961	6	LH200xLH172	8	4
961	6	LH195xNC258	9	4
961	10	DK683	10	4
961	47	P3394	1	5
961	37	FR1064xLH185	2	5
961	26	P3163	3	5
961	20	LH195xLH212	4	5
961	31	LH195xLH210	5	5
961	27	LH235xLH185	6	5
961	21	LH200xLH216	7	5
961	27	LH200xLH172	8	5
961	6	LH195xNC258	9	5
961	6	DK683	10	5
961	27	P3394	1	6
961	27	FR1064xLH185	2	6
961	13	P3163	3	6
961	19	LH195xLH212	4	6
961	17	LH195xLH210	5	6
961	9	LH235xLH185	6	6
961	18	LH200xLH216	7	6
961	4	LH200xLH172	8	6
961	5	LH195xNC258	9	6
961	2	DK683	10	6
961	23	P3394	1	7
961	15	FR1064xLH185	2	7
961	4	P3163	3	7
961	8	LH195xLH212	4	7
961	10	LH195xLH210	5	7
961	8	LH235xLH185	6	7
961	7	LH200xLH216	7	7
961	6	LH200xLH172	8	7
961	2	LH195xNC258	9	7
961	4	DK683	10	7
961	100	P3394	1	8
961	100	FR1064xLH185	2	8
961	48	P3163	3	8
961	81	LH195xLH212	4	8
961	82	LH195xLH210	5	8
961	83	LH235xLH185	6	8
961	92	LH200xLH216	7	8
961	82	LH200xLH172	8	8
961	13	LH195xNC258	9	8
961	18	DK683	10	8
961	19	P3394	1	9
961	14	FR1064xLH185	2	9
961	7	P3163	3	9
961	5	LH195xLH212	4	9
961	3	LH195xLH210	5	9
961	6	LH235xLH185	6	9
961	3	LH200xLH216	7	9
961	2	LH200xLH172	8	9
961	1	LH195xNC258	9	9
961	1	DK683	10	9
961	24	P3394	1	10
961	24	FR1064xLH185	2	10
961	8	P3163	3	10
961	21	LH195xLH212	4	10
961	22	LH195xLH210	5	10
961	16	LH235xLH185	6	10
961	16	LH200xLH216	7	10
961	19	LH200xLH172	8	10
961	19	LH195xNC258	9	10
961	5	DK683	10	10
961	62	P3394	1	11
961	33	FR1064xLH185	2	11
961	35	P3163	3	11
961	26	LH195xLH212	4	11
961	19	LH195xLH210	5	11
961	27	LH235xLH185	6	11
961	20	LH200xLH216	7	11
961	34	LH200xLH172	8	11
961	18	LH195xNC258	9	11
961	9	DK683	10	11
;

title 'Gray leaf spot, South 96, Final Y';
%InteractE(data=s96,blk=loc,trt=gen,
		response=x, noblks=11,notrts=10, nocols=5,
		print=no,spearman=no,odd=Yes);
run;
quit;
